#!/usr/local/bin/python
import sys
# sys.path.append('/Users/jore/courses/NIMBUS/RESEARCH/CPS_TYPES/cps_units/')
import unittest
from detect_physical_unit_inconsistencies import CPSUnitsChecker
from unit_error_types import UnitErrorTypes
from unit_error import UnitError
import os

global_debug = False
global_debug_verbose = False
global_debug_AST = False

class TestStringMethods(unittest.TestCase):

        
    def test_function_return_0(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        cps_unit_checker.debug_scope = False
        dump_file  = './dump_files_for_tests/test_it_function_return_0.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        units_for_f1 = []
        # TEST THAT UNITS ARE ASSIGNED TO FUNCTION
        for tw in cps_unit_checker.all_tree_walkers:
            so = tw.symbol_helper.function_dictionary['scopeObject']
            if so.function:
                if so.function.name == 'f1':
                    units_for_f1 = so.function.return_units
        self.assertEquals(units_for_f1, [{'meter': 1}], 'Incorrect units returned for function: f1 .  Expected [{\'meter\':1}], received %s' % units_for_f1)
        # TEST THAT UNITS ARE RECEIVED TO FUNCTION
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][12]['units']
                
        my_oracle = [{'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        

            
    def test_function_return_1(self):
        ''' x SHOULD END UP M/S, but so far THERE'S NO MECHANISM FOR PASSING UNITS IN TO A FUNCTION
        '''
        cps_unit_checker = CPSUnitsChecker()
        dump_file  = './dump_files_for_tests/test_it_function_return_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for tw in cps_unit_checker.all_tree_walkers:
            so = tw.symbol_helper.function_dictionary['scopeObject']
            if so.function:
                if so.function.name == 'f1':
                    units_for_f1 = so.function.return_units

    def test_comparisons_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST= False
        dump_file  = './dump_files_for_tests/test_it_comparisons_1.cpp.dump'
        source_file  = './dump_files_for_tests/test_it_comparisons_1.cpp'
        cps_unit_checker.main_run_check(dump_file, source_file)
        e = cps_unit_checker.errors[0]
        # ORACLES
        token_left_units_oracle = [{'meter': 1}]
        token_right_units_oracle = [{'second': -1, 'meter': 1}]
        # ASSERTIONS
        self.assertEqual(e.token.str, '>')
        self.assertEqual(e.token_left.units, token_left_units_oracle)
        self.assertEqual(e.token_right.units, token_right_units_oracle)

    def test_logical_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_logical_1.cpp.dump'
        source_file  = './dump_files_for_tests/test_it_logical_1.cpp'
        cps_unit_checker.main_run_check(dump_file, source_file)
        # TEST 1
        e = cps_unit_checker.errors[0]
        # ORACLES
        token_right_units_oracle = [{'meter': 1}]
        # ASSERTIONS
        self.assertEqual(e.linenr, 13)
        self.assertEqual(e.token.str, '&&')
        self.assertEqual(e.token_right.units, token_right_units_oracle)
        # TEST 2
        e = cps_unit_checker.errors[1]
        # ORACLES
        token_left_units_oracle = [{'meter': 1}]
        # ASSERTIONS
        self.assertEqual(e.linenr, 18)
        self.assertEqual(e.token.str, '||')
        self.assertEqual(e.token_left.units, token_left_units_oracle)

    def test_abs_0(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_abs.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)

    
    def test_abs_namespace_std_0(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_abs_namespace_std.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)

    def test_abs_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_abs_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 't'
        var_linenr = 9
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        my_oracle = [{'second': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_abs_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_abs_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 's'
        var_linenr =11 
        my_oracle = [{'meter': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_multiplication_assignment_in_multi_configurations_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiplication_assignment_in_multi_configurations.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'a_geometry_msgs_Accel.linear.x'
        var_linenr = 19 
        my_oracle = [{'second': -4, 'meter': 2}] 
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_unit_propagation_by_multiplication_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        #cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_unit_propagation_by_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][14]['units']
        my_oracle = [{'second': -4, 'meter': 2}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_unit_propagation_by_division_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        #cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_unit_propagation_by_division_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][14]['units']
        my_oracle = None
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_mulitple_units_assigned(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiple_units_assigned_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        expected_errors = ["test_it_multiple_units_assigned_1.cpp : 11  MULTIPLE UNITS BY ASSIGNMENT: [{'second': -1, 'meter': 1}, {'second': -2, 'meter': 2}]"]
        # self.assertListEqual([e['error_msg'] for e in cps_unit_checker.errors], expected_errors)
        # TEST QUANTITY OF ERRORS
        self.assertEqual(1, len(cps_unit_checker.errors))
        # TEST TyPE OF ERROR
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        # TEST VALUE OF ERROR
        var_name = 'a_geometry_msgs_Accel.linear.x'
        var_linenr =11
        my_oracle = [{'second': -2, 'meter': 1}, {'second': -4, 'meter': 2}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_known_functions_sqrt_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][12]['units']
        my_oracle = [{'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_known_functions_sqrt_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][12]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_known_functions_sqrt_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][12]['units']
        my_oracle = None
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_known_functions_sqrt_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][14]['units']
        my_oracle = [{'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_known_functions_sqrt_5(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_5.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][14]['units']
        my_oracle = [{'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_known_functions_sqrt_half_units(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_sqrt_half_units.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'x' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['x'][11]['units']
        my_oracle = [{'meter': 0.5}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_known_functions_atan2_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_atan2_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][7]['units']
        my_oracle = [{'radian': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_known_functions_atan2_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        #cps_unit_checker.debug_verbose = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_atan2_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][8]['units']
        my_oracle = [{'radian': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_toSec(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_toSec_0.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'duration' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['duration'][7]['units']
        my_oracle = [{'second': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'second' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['second'][9]['units']
        my_oracle = [{'second': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_float_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_float_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict and 11 in s.var_ordered_dict['f']:
                    actual_units = s.var_ordered_dict['f'][11]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_float_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_float_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict and 11 in s.var_ordered_dict['f']:
                    actual_units = s.var_ordered_dict['f'][11]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_float_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_float_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][12]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_float_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_float_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][13]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_float_5(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_float_5.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if 'f' in s.var_ordered_dict:
                    actual_units = s.var_ordered_dict['f'][11]['units']
        my_oracle = [{'second': -1, 'meter': 1}]
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_pow_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_pow_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr =10 
        my_oracle = [{'meter': 4}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_pow_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_pow_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 11 
        my_oracle = [{'meter': 4}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_pow_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_pow_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 10
        my_oracle = [{'meter': 4}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:%s Expected: %s received %s' %  (var_name, my_oracle, actual_units))

    def test_floor_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_floor_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 's'
        var_linenr = 8 
        my_oracle = [{'meter': 1, 'second':-1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_ceil_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_ceil_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 's'
        var_linenr = 8 
        my_oracle = [{'meter': 1, 'second':-1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_acos_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_acos_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 7 
        my_oracle = [{'radian': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_asin_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_asin_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 7 
        my_oracle = [{'radian': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_atan_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_known_function_atan_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 7 
        my_oracle = [{'radian': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_ternary_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_ternary_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 9 
        my_oracle = [{'second': -1, 'meter': 1}, {'second': -1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_function_args_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_function_args_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # actual_units = None
        f = cps_unit_checker.current_configuration.functions[0].arg_units
        self.assertEqual(f[0][0]['linenr'], 13)
        self.assertEqual(f[0][0]['units'], [{'meter': 1}])


    def test_function_args_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_function_args_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        for f in cps_unit_checker.current_configuration.functions[0].arg_units:
            self.assertEqual(f[0]['linenr'], 13)
            self.assertEqual(f[0]['units'], [{'meter': 1}])

    def test_function_args_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_function_args_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        my_oracle_1 = 4
        my_oracle_2 = [{'meter': 1}]
        actual_units = None
        all_units_list = cps_unit_checker.current_configuration.functions[0].arg_units
        self.assertEqual(len(all_units_list), my_oracle_1)
        for u in all_units_list:
            self.assertEqual(u[0]['units'], my_oracle_2)

    def test_function_args_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_function_args_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        my_oracle = [{'meter': 1}]
        actual_units = None
        for f in cps_unit_checker.current_configuration.functions:
            for arg_u in f.arg_units:
                for arg_use_on_line in arg_u:
                    self.assertEqual(arg_use_on_line['units'], my_oracle)

    def test_function_args_5(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_function_args_5.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        my_oracle_1 = [{'meter': 1}]
        my_oracle_2 = 15
        f = cps_unit_checker.current_configuration.functions[0]
        self.assertEqual(f.arg_units[0][0]['units'], my_oracle_1)
        self.assertEqual(f.arg_units[0][0]['linenr'], my_oracle_2)

    def test_division_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_division_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr = 9 
        my_oracle = [{'meter': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_division_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_division_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr = 9 
        my_oracle = [{'meter': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_division_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_division_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr = 9 
        my_oracle = [{'second': 2, 'meter': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))


    def test_division_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_division_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr =10
        my_oracle = [{'second': 2}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_logical_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_logical_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    def test_error_type_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_error_return_type_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.current_file_under_analysis = dump_file
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))

    def test_laser_scan_range_size_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_laser_scan_range_count_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr = 7
        my_oracle = None
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(0, len(cps_unit_checker.errors))

    def test_laser_scan_range_size_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_laser_scan_range_count_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'x'
        var_linenr = 7
        my_oracle = [{'meter':1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_ros_duration_isZero_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_ros_duration_isZero_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 't'
        var_linenr = 6
        my_oracle = None
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_ros_duration_isZero_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_ros_duration_isZero_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 't'
        var_linenr = 6
        my_oracle = [{'second':1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    def test_ros_header_include_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/src/test_it_header_include_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(2, len(cps_unit_checker.errors))

    def test_ros_header_include_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/src/test_it_header_include_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(3, len(cps_unit_checker.errors))
        # WEAKER - SOMETHING STOCASTIC IS HAPPENING
        e = cps_unit_checker.errors[0]
        self.assertEqual(7, e.linenr)
        self.assertEqual('./dump_files_for_tests/src/../include/test_it_header_include_2.h', e.get_file_URI_where_error_occured())
        e = cps_unit_checker.errors[1]
        self.assertEqual(5, e.linenr)
        self.assertEqual('./dump_files_for_tests/src/test_it_header_include_2.cpp', e.get_file_URI_where_error_occured())

    #   DON'T ASSIGN UNITS TO ARRAYS WHEN array.empty() IS CALLED 
    def test_laser_range_empty_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_range_empty_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    #   DON'T ASSIGN UNITS TO ARRAYS WHEN time.isZero() IS CALLED 
    def test_ros_isZero_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_ros_isZero_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    #   DON'T ASSIGN UNITS DURING x = y = z = 0
    def test_multiple_initialization_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiple_initialization.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    #   WEAKEN ASSIGNMENT WHEN MULTIPLIED BY A CONSTANT (INT)
    def test_it_multiplication_with_constant_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiplication_with_constant_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # self.assertEqual(0, len(cps_unit_checker.errors))
        var_name = 'f'
        var_linenr = 9
        my_oracle = [{'second':-1}]
        actual_units = None
        is_unit_propagation_based_on_constants = False
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
                    is_unit_propagation_based_on_constants = s.var_ordered_dict[var_name][var_linenr]['is_unit_propagation_based_on_constants']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertTrue(is_unit_propagation_based_on_constants, 'Unit inference should be weakened by constant interaction, but is still strong.')

    #   WEAKEN ASSIGNMENT WHEN MULTIPLIED BY A CONSTANT  (FLOAT)
    def test_it_multiplication_with_constant_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiplication_with_constant_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # self.assertEqual(0, len(cps_unit_checker.errors))
        var_name = 'f'
        var_linenr = 9
        my_oracle = [{'second':-1}]
        actual_units = None
        is_unit_propagation_based_on_constants = False
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
                    is_unit_propagation_based_on_constants = s.var_ordered_dict[var_name][var_linenr]['is_unit_propagation_based_on_constants']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertTrue(is_unit_propagation_based_on_constants, 'Unit inference should be weakened by constant interaction, but is still strong.')

    #   WEAKEN ASSIGNMENT WHEN MULTIPLIED BY A CONSTANT  (FLOAT)
    def test_it_operator_with_unknown_variable_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_operator_with_unknown_variable_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # self.assertEqual(0, len(cps_unit_checker.errors))
        var_name = 'f'
        var_linenr = 10 
        my_oracle = [{'second':-1}]
        actual_units = None
        is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
                    is_unit_propagation_based_on_unknown_variable = s.var_ordered_dict[var_name][var_linenr]['is_unit_propagation_based_on_unknown_variable']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertTrue(is_unit_propagation_based_on_unknown_variable, 'Unit inference should be weakened by unknown variable interaction, but is still strong.')

    # WEAKEN ERROR WHEN MULTIPLIED BY A CONSTANT 
    def test_it_operator_with_unknown_variable_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_operator_with_unknown_variable_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(2, len(cps_unit_checker.errors))
        for e in cps_unit_checker.errors:
            self.assertTrue(e.is_warning, 'Should be a warning but is not marked as such')

    # WEAKEN ERROR WHEN MULTIPLIED BY A CONSTANT 
    def test_it_operator_with_unknown_variable_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_operator_with_unknown_variable_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(2, len(cps_unit_checker.errors))

    # PROPAGATION ACROSS MIN MAX
    def test_it_min_max_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_min_max_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 7
        my_oracle = [{'second': -1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(0, len(cps_unit_checker.errors))

    # PROPAGATION ACROSS MIN MAX
    def test_it_min_max_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_min_max_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 8
        my_oracle = [{'second': -1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(0, len(cps_unit_checker.errors))


    # PROPAGATION ACROSS MIN MAX
    def test_it_min_max_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_min_max_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 8
        my_oracle = [{'second': -1}, {'second': -1, 'meter': 1}]
        actual_units = None
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[0].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)



    # PROPAGATION ACROSS MIN MAX
    def test_it_min_max_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_min_max_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 9
        my_oracle = [{'second': -1 }, {'second': -1, 'meter': 1}]
        actual_units = None
        is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[0].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)



    # PROTECTION AGAINST MULTILINE
    def test_it_multiline_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_multiline_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'f'
        var_linenr = 25
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'second': -2.0, 'meter': 2.0}, {'second': -3.0, 'meter': 3.0}, {'second': -2.0, 'meter': 1.0}, {'second': -3.0, 'meter': 2.0}, {'second': -4.0, 'meter': 3.0}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))

    # KNOW FUNCTION quatToRPY
    def test_it_quatToRPY_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quatToRPY_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'tw.linear.x'
        var_linenr = 17
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian': 1.0}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)


    # WEAK INFERENCE WARNING
    def test_it_weak_inference_multiplication_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_weak_inference_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference: %s warning:%s ' % (e.var_name, str(e.is_warning))
        var_name = 'tw.linear.x'
        var_linenr = 19
        my_oracle = [{'second': -1.0, 'meter': 1.0}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(0, len(cps_unit_checker.errors))

    # WEAK INFERENCE WARNING
    def test_it_weak_inference_multiplication_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_weak_inference_multiplication_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'tw.linear.x'
        var_linenr = 22
        my_oracle = [{'second': -1.0, 'meter': 1.0}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(0, len(cps_unit_checker.errors))

    # STRONG INFERENCE BECAUSE ADDITION
    def test_it_weak_inference_addition_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_weak_inference_addition_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        var_name = 'tw.linear.x'
        var_linenr = 22
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(2, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[0].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)
        var_name = 'tw.linear.y'
        var_linenr = 23
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertTrue(cps_unit_checker.errors[1].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[1].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[1].is_warning)

    # STRONG INFERENCE BECAUSE ADDITION - SWAPPED OPERAND ORDER
    def test_it_weak_inference_addition_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_weak_inference_addition_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        var_name = 'tw.linear.x'
        var_linenr = 22
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(2, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[0].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)
        var_name = 'tw.linear.y'
        var_linenr = 23
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertTrue(cps_unit_checker.errors[1].was_assigned_mutiple_units)
        self.assertFalse(cps_unit_checker.errors[1].is_unit_propagation_based_on_unknown_variable)
        self.assertFalse(cps_unit_checker.errors[1].is_warning)

    # STRONG INFERENCE BECAUSE ADDITION - SWAPPED OPERAND ORDER
    def test_it_weak_inference_addition_3 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_weak_inference_addition_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        var_name = 'tw.linear.x'
        var_linenr = 22
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'radian':1.0}, {'second':1.}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(2, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)
        self.assertTrue(cps_unit_checker.errors[0].is_unit_propagation_based_on_unknown_variable)
        self.assertTrue(cps_unit_checker.errors[0].is_warning)


    # ADDITION STAND ALONE ERROR FOR ADDITION OF INCOMPATIBLE UNITS - STRONG
    def test_it_addition_without_assignment_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_addition_without_assignment_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # ADDITION STAND ALONE ERROR FOR ADDITION OF INCOMPATIBLE UNITS - WEAK UNKNOWN VARIABLE
    def test_it_addition_without_assignment_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_addition_without_assignment_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertTrue(cps_unit_checker.errors[0].is_warning)

    # ADDITION STAND ALONE ERROR FOR ADDITION OF INCOMPATIBLE UNITS - WEAK CONSTANT
    def test_it_addition_without_assignment_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_addition_without_assignment_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertTrue(cps_unit_checker.errors[0].is_warning)

    # ADDITION STAND ALONE ERROR FOR SUBTRACTION OF INCOMPATIBLE UNITS - STRONG CONSTANT
    def test_it_addition_without_assignment_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_addition_without_assignment_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)


    # ADDITION OF RADIANS
    def test_it_radian_addition_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_addition_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(2, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[1].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[1].is_warning)


    # ADDITION OF RADIANS
    def test_it_radian_addition_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_addition_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # MULTIPLICATION OF RADIANS
    def test_it_radian_multiplication_1(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # MULTIPLICATION OF RADIANS 2
    def test_it_radian_multiplication_2(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_multiplication_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # MULTIPLICATION OF RADIANS
    def test_it_radian_multiplication_3(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_multiplication_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # MULTIPLICATION OF RADIANS 2
    def test_it_radian_multiplication_4(self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_multiplication_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)


    # getXYZ 
    def test_it_getXYZ_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_getXYZ_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        self.assertEqual(0, len(cps_unit_checker.errors))

    # getXYZ 
    def test_it_getXYZ_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_getXYZ_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        var_name = 'tw.linear.x'
        var_linenr = 10
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'meter':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)


    # getXYZ 
    def test_it_getXYZ_3 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_getXYZ_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        var_name = 'tw.linear.x'
        var_linenr = 10
        my_oracle = [{'second': -1.0, 'meter': 1.0}, {'quaternion':1}]
        actual_units = None
        # is_unit_propagation_based_on_unknown_variable = False
        for s in cps_unit_checker.current_configuration.scopes:
            # for v in s.var_ordered_dict:
                # print v
            if s.className == 'main':
                if var_name in s.var_ordered_dict and var_linenr in s.var_ordered_dict[var_name]:
                    actual_units = s.var_ordered_dict[var_name][var_linenr]['units']
        self.assertEquals(actual_units, my_oracle, 'Incorrect units assigned to symbol:x Expected: %s received %s' %  (my_oracle, actual_units))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertTrue(cps_unit_checker.errors[0].was_assigned_mutiple_units)

    # getXYZ 
    def test_it_getXYZ_4 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_getXYZ_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        # for e in cps_unit_checker.errors:
            # print '\nweak inference addition : %s warning:%s ' % (e.var_name, str(e.is_warning))
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # getXYZ 
    def test_it_getXYZ_5 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_getXYZ_5.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # QUATERNION ADDITION 1
    def test_it_quaternion_addition_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_addition_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(2, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)
        self.assertEqual(UnitErrorTypes.ADDITION_OF_INCOMPATIBLE_UNITS, cps_unit_checker.errors[1].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[1].is_warning)


    # QUATERNION ADDITION 2
    def test_it_quaternion_addition_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_addition_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)


    # QUATERNION ADDITION 3
    def test_it_quaternion_addition_3 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_addition_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION ADDITION 4
    def test_it_quaternion_addition_4 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_addition_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION MULTIPLICATION 1
    def test_it_quaternion_multiplication_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION MULTIPLICATION 2
    def test_it_quaternion_multiplication_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_multiplication_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION MULTIPLICATION 3
    def test_it_quaternion_multiplication_3 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_multiplication_3.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION MULTIPLICATION 4
    def test_it_quaternion_multiplication_4 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_multiplication_4.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # QUATERNION MULTIPLICATION CLOSURE
    def test_it_quaternion_closed_under_multiplication_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_closed_under_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))


    # QUATERNION MULTIPLICATION CLOSURE
    def test_it_quaternion_closed_under_multiplication_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_quaternion_closed_under_multiplication_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)

    # RADIAN MULTIPLICATION CLOSURE
    def test_it_radian_closed_under_multiplication_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_closed_under_multiplication_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # RADIAN MULTIPLICATION CLOSURE
    def test_it_radian_closed_under_multiplication_2 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_radian_closed_under_multiplication_2.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(1, len(cps_unit_checker.errors))
        self.assertEqual(UnitErrorTypes.VARIABLE_MULTIPLE_UNITS, cps_unit_checker.errors[0].ERROR_TYPE)
        self.assertFalse(cps_unit_checker.errors[0].is_warning)


    # dt Heuristic
    def test_it_dt_heuristic (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_dt_heuristic_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # dt Heuristic
    def test_it_plus_equals_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_plus_equals_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

    # dt Heuristic
    def test_it_range_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_range_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))


    # same named argument in interface scope bug
    def test_it_scope_bug_1 (self):
        cps_unit_checker = CPSUnitsChecker()
        cps_unit_checker.debug = False
        cps_unit_checker.debug_print_AST = False
        dump_file  = './dump_files_for_tests/test_it_cppcheck_scope_bug_at_argument_1.cpp.dump'
        source_file  = dump_file.replace('.dump','')
        cps_unit_checker.main_run_check(dump_file, source_file)
        self.assertEqual(0, len(cps_unit_checker.errors))

if __name__ == '__main__':
    unittest.main()




class UnitErrorTypes:
    """ DEFINES TYPES OF UNIT ERRORS
    """
    VARIABLE_MULTIPLE_UNITS = 0
    COMPARISON_INCOMPATIBLE_UNITS = 1
    VARIABLE_BECOME_UNITLESS = 2
    FUNCTION_CALLED_WITH_DIFFERENT_UNIT_ARGUMENTS = 3
    VARIABLE_WITH_UNUSUAL_UNITS = 4
    ADDITION_OF_INCOMPATIBLE_UNITS = 5
    LOGICAL_OPERATOR_USED_ON_UNITS = 6
    UNIT_SMELL = 7

    SEVERITY = [                    # MAPPED DIRECTLY FROM ROS VERBOSITY LEVELS
                    'DEBUG',
                    'INFO',
                    'WARN',
                    'ERROR',
                    'FATAL',
                   ]
    ERR_TYPE_NAMES = [              # MUST BE SAME INDEX AS ERRORS LISTED ABOVE
                        'MULTIPLE UNITS FOR VARIABLE',
                        'COMPARISON OF INCOMPATIBLE UNITS',
                        'VARIABLE BECAME UNITLESS',
                        'FUNCTION CALLED WITH DIFFERENT UNIT ARGUMENTS',
                        'VARIABLE WITH UNUSUAL UNITS',
                        'ADDITION OF INCOMPATIBLE UNITS',
                        'LOGICAL OPERATOR USED ON UNITS',
                        'UNIT SMELL',
                        ]

    def __init__(self):
        pass

    def get_err_short_discription(self, num):
        return self.ERR_TYPE_NAMES[num]

s = '''
VARIABLE MULTIPLE UNITS
COMPARISON INCOMPATIBLE UNITS
VARIABLE BECOME UNITLESS
FUNCTION CALLED WITH DIFFERENT UNIT ARGUMENTS
VARIABLE WITH UNUSUAL UNITS
ADDITION OF INCOMPATIBLE UNITS
LOGICAL OPERATOR USED ON UNITS
UNIT_SMELL
'''



import copy
import re

class SymbolHelper:
    ''' HELPS FIND DEFINITIONS OF SYMBOLS AND DECORATES CPPCHECK SYMBOL TABLE
    '''

    def __init__(self, function_dictionary):
        self.SHOULD_GIVE_VECTOR3_METERS = False
        self.function_dictionary = function_dictionary
        self.should_use_variable_names = False
        self.should_ignore_time_and_math = False
        self.should_use_dt_heuristic = False
        self.is_weak_inference = False
        self.debug = False
        self.ros_unit_dictionary = {}
        self.list_of_vetted_units = []
        self.cps_unit_checker = None   # DEPENDENCY INJECTION.  SEE MARTIN FOWLER
        self.initialize_ros_unit_dictionary()
        self.initialize_function_dictionary()
        self.initialize_list_of_vetted_units()
        self.should_match_on_heuristic_variable_names = False #USED TO COMPARE MORE NAIVE METHOD OF ASSIGNING UNITS BY SUBSTRING, LIKE 'linear' = {'meter':1, 'second':-1}
        self.debug_missed_class_names_output_file = '2016_12_04_all_missed_class_name_lookups.txt'
        self.debug_log_missed_class_names = True
        self.weak_inference_classes = ['sensor_msgs::JointState']
        # THESE DIMENSIONLESS UNITS ACT AS A '1' DURTION DIMENSION OPERATIONS, EXEPCT FOR ADDITION.  
        # EXAMPLES    radians + meters NOT OK.  radians * meters = meters.  quaternion * quaternion = quaternion
        self.dimensionless_units = [{'radian':1.}, {'quaternion':1.}]
        self.dimensionless_units_as_lists = [[{'radian':1.}], [{'quaternion':1.}]]


    # STACK OVERFLOW
    def rreplace(self, s, old, new, occurance):
        li = s.rsplit(old, occurance)
        return new.join(li)


    def sanitize_class_name(self, class_name):
        ''' TAKES A VARIABLE NAME AND STRIPTS EXTRA CLASS INFORMATION FROM FRONT AND BACK
            input: class_name   'constPtr<tf2::Transform::iterator>'  <-- LOL
            output: str         'trf2::Transform'
            '''
        is_some_change = True
        while (is_some_change):# SOME OF THESE CAN BE CASCADED IN DIFFERENT ORDERS
            is_some_change = False

            # SIMPLIFY ITERATORS
            if str(class_name).endswith('::iterator'):
                class_name = self.rreplace(class_name, '::iterator', '', 1)
                is_some_change = True
            if str(class_name).endswith('::const_iterator'):
                class_name = self.rreplace(class_name, '::const_iterator', '', 1)
                is_some_change = True


            # SIMPLIFY REFERNCES
            if str(class_name).endswith('&'):
                class_name = class_name.replace('&', '')
                is_some_change = True
            # SIMPLIFY CONST
            if str(class_name).endswith('const'):
                class_name = self.rreplace(class_name, 'const', '', 1)
                is_some_change = True
            # SIMPLIFY CONSTANT POINTERS (REFRENCES)
            if str(class_name).endswith('ConstPtr'):
                class_name = class_name.replace('ConstPtr', '')
                is_some_change = True
            # SIMPLIFY POINTERS
            if str(class_name).endswith('::Ptr'):
                class_name = self.rreplace(class_name, '::Ptr', '', 1)
                is_some_change = True
            if str(class_name).endswith('Ptr'):
                class_name = self.rreplace(class_name, 'Ptr', '', 1)
                is_some_change = True
            # SIMPLIFY BOOST POINTERS
            if str(class_name).startswith('boost::shared_ptr<') and str(class_name).endswith('>'):
                class_name = class_name.replace('boost::shared_ptr<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('shared_ptr<') and str(class_name).endswith('>'):
                class_name = class_name.replace('shared_ptr<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('boost::scoped_ptr<') and str(class_name).endswith('>'):
                class_name = class_name.replace('boost::scoped_ptr<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('scoped_ptr<') and str(class_name).endswith('>'):
                class_name = class_name.replace('scoped_ptr<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('tf::MessageFilter<') and str(class_name).endswith('>'):
                class_name = class_name.replace('tf::MessageFilter<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('tf2_ros::MessageFilter<') and str(class_name).endswith('>'):
                class_name = class_name.replace('tf_ros::MessageFilter<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('MessageFilter<') and str(class_name).endswith('>'):
                class_name = class_name.replace('MessageFilter<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True

            # SIMPLIFY VECTORS
            if str(class_name).startswith('std::vector<') and str(class_name).endswith('>'):
                class_name = class_name.replace('std::vector<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('conststd::vector<') and str(class_name).endswith('>'):
                class_name = class_name.replace('std::vector<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('vector<') and str(class_name).endswith('>'):
                class_name = class_name.replace('vector<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('std::deque<') and str(class_name).endswith('>'):
                class_name = class_name.replace('std::vector<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            if str(class_name).startswith('deque<') and str(class_name).endswith('>'):
                class_name = class_name.replace('vector<', '')
                class_name = self.rreplace(class_name, '>', '', 1)
                is_some_change = True
            # SIMPLIFY POINTERS
            if str(class_name).endswith('*'):
                class_name = self.rreplace(class_name, '*', '', 1)
                is_some_change = True
            # SIMPLIFY DANGLING ::
            if str(class_name).endswith('::'):
                class_name = self.rreplace(class_name, '::', '', 1)
                is_some_change = True
        # SANITIZED VERSION
        return class_name



    def find_units_for_variable(self, token):
        ''' input:  token       a cppcheckdata token that contains a variable 
            returns: a dictionary of unit types
            '''
        # HOLDER FOR RETURNED UNITS, IF FOUND
        my_return_units = {}

        # ASSUME FALSE
        self.is_weak_inference = False

        # TRAVERSE CPPCHECK SYMBOL LINK STRUCTURE TO VARIABLE DEFINITION TO EXTRACT TYPE
        raw_class_name = self.find_variable_type(token.variable)

        # REMOVE EXTRA, UNNECESSARY INFORMATION FROM FRONT AND BACK
        class_name = self.sanitize_class_name(raw_class_name)

        # IF CLASS IS A ROS MESSAGE
        if class_name in self.ros_unit_dictionary:
            # FIND POSSIBLE ATTRIBUTES
            possible_attributes_as_list = self.find_compound_variable_name_for_variable_token(token).split(".")
            if possible_attributes_as_list:
                # STARTED AS SPECIAL CASE FOR LaserScan
                # CORNER CASES WHEN WE DONT' ASSIGN UNITS
                # if (class_name == 'sensor_msgs::LaserScan') and 'size' in possible_attributes_as_list:
                if 'size' == possible_attributes_as_list[-1]:
                    return {}
                if 'empty' == possible_attributes_as_list[-1]:
                    return {}
                if 'isZero' ==  possible_attributes_as_list[-1]:
                    return {}
                # THE USUAL EXPECTED CASE 
                for p in possible_attributes_as_list:
                    if p in self.ros_unit_dictionary[class_name]:
                        my_return_units = self.ros_unit_dictionary[class_name][p]
            if not self.should_ignore_time_and_math:
                if class_name in [ 'ros::Time', 'ros::Duration', 'std::vector<ros::Duration>']:
                    my_return_units = self.ros_unit_dictionary[class_name]['secs']
            if my_return_units:
                if self.cps_unit_checker:
                    if self.cps_unit_checker.SHOULD_FIND_ALL_UNITS:
                        self.cps_unit_checker.add_class_and_units_to_all_list(class_name, my_return_units)
                        if not class_name in self.cps_unit_checker.all_classes_and_units_for_this_file_as_dict:
                            self.cps_unit_checker.all_classes_and_units_for_this_file_as_dict[class_name] = my_return_units
                        else:
                            pass  # WE DON'T CASE BECAUSE WE'RE ONLY TRACKING UNIQUE UNITS, NOT COUNTS
                # CHECK FOR WEAK INFERENCE CLASSES, SUCH AT JOINT_STATE BECAUSE THAT COULD BE r/s or m/s
                if class_name in self.weak_inference_classes:
                    self.is_weak_inference = True
                # with open(self.debug_log_all_variable_unit_assignments, 'a') as f:
                # with open(self.debug_missed_class_names_output_file, 'a') as f:
                    # f.write(class_name + ", " + token.str + ", " + str(my_return_units) + '\n')
                return my_return_units
        # RETURN STATEMENTS ABOVE IMPLY THIS POINT MIGHT NOT BE REACHED

        # OPTIONAL NAME MATCHING HEURISTIC
        if self.should_match_on_heuristic_variable_names:
            my_return_units = self.get_units_by_name_heuristic(token, class_name)
            # if self.should_use_dt_heuristic and not self.should_ignore_time_and_math:
                # if token.str.lower() == 'dt':
                    # return self.ros_unit_dictionary['dt']['dt']

        if not my_return_units:  
            # REACHNG HERE MEANS WE FOUND NOTHING
            if self.debug_log_missed_class_names:
                pass
                # with open(self.debug_missed_class_names_output_file, 'a') as f:
                    # f.write(class_name + ", " + token.str + '\n')

        return my_return_units

    def get_units_by_name_heuristic(self, token, class_name):
        ''' GIVEN A TOKEN STRING, ATTEMPT TO FIND UNITS BASED ON VARIABLE NAMING HEURISTICS
            input: raw token string, class name as string
            output: unit dictionary (dict of dicts) if found, otherwise empty dict
            '''
        my_return_units = {}
        possible_attributes_as_list = self.find_compound_variable_name_for_variable_token(token).split(".")
        # DONT' ASSIGN UNITS FOR CERTAIN VARIABLE TYPES, LIKE BOOL
        if 'bool' in class_name:
            return {}

        for member_name in possible_attributes_as_list:
            variable_name = str(member_name).upper()
            if variable_name == 'DT':   # dT TIME HEURISTIC  CASE INSENSITIVE
                return {'second':1.}
            if 'TIME' == variable_name :   
                return {'second':1.}
            if 'DURATION' == variable_name :   
                return {'second':1.}
            if 'FORCE' == variable_name:   #  CASE INSENSITIVE
                return {'kilogram':1., 'meter': 1., 'second':-2.}
            # if 'TORQ' in variable_name:   # torq CASE INSENSITIVE
                # return {'kilogram':1., 'meter': 2., 'second':-2.}
            # if 'POSITION' in variable_name:
                # return {'meter':1.}
            # if 'RANGE' in variable_name:
                # return {'meter':1.}
            # if 'VEL' in variable_name:   # vel VELOSITY  CASE INSENSITIVE
                # return {'meter':1., 'second':-1.}
            # 'stamp':{'second':1.}}
            

            


    def find_compound_variable_name_for_variable_token(self, token):
        ''' input: a variable token from cppcheckdata
            returns: string containing the compound variable name.  ex:  'my_var_.linear.x'
            '''
        if not token.variable:
            raise ValueError('received a non variable token for tokenid:%s str:%s' % (token.Id, token.str))
        if token.astParent.str != '.':
            return token.str
        compound_variable_root_token = token.astParent
        while compound_variable_root_token.astParent and compound_variable_root_token.astParent.str == '.':
            compound_variable_root_token = compound_variable_root_token.astParent
        return self.recursively_visit(compound_variable_root_token)
            
    def recursively_visit(self, token):
        ''' input: a cppcheckdata token
            returns: string aggregation of tokens under the root
            '''
        my_return_string = ''
        if token.astOperand1: my_return_string += self.recursively_visit(token.astOperand1)
        my_return_string += token.str 
        if token.astOperand2:  my_return_string += self.recursively_visit(token.astOperand2)
        return my_return_string


    #def get_unit_dict_given_class_and_property_as_strings(self, class_name, property_name):
        #''' input:  class name (ex: 'sensor_msgs::BatteryState')
                    #property_name (ex: 'cell_voltage')
            #returns: dict containing units, if found, otherwise empty dict  (ex: '{'volt': 1})
            #'''
        #if self.should_match_on_heuristic_variable_names:
            #return self.get_unit_dict_STRING_MATCH_HEURISTIC()

        #nope
        #print class_name
        #foo
        #class_name = class_name.replace('::ConstPtr&', '')

        #if class_name not in self.ros_unit_dictionary:
            #return {}
        #if property_name not in self.ros_unit_dictionary[class_name]:
            #return {}
        #return self.ros_unit_dictionary[class_name][property_name]

    def initialize_function_dictionary(self):
        #self.functionDictionary
        pass

    def initialize_ros_unit_dictionary(self):


        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  WRENCH
        twist_dict                                      = {'linear': {'meter': 1., 'second': -1.}, 
                                                            'angular': {'second': -1.}}
        accel_dict                                      = {'linear': {'meter': 1., 'second': -2.}, 
                                                            'angular': {'second': -2.}}
                                                            #'covariance'} todo
        wrench_dict                                     = {'force': {'kilogram':1., 'meter': 1., 'second':-2.}, 
                                                            'torque': {'kilogram':1., 'meter': 2., 'second':-2.}}
        inertia_dict                                    = {'m': {'kilogram': 1.},
                                                            'com': {'meter':1.},
                                                            'ixx': {'kilogram':1., 'meter':-2. },
                                                            'ixy': {'kilogram':1., 'meter':-2. },
                                                            'ixz': {'kilogram':1., 'meter':-2. },
                                                            'iyy': {'kilogram':1., 'meter':-2. },
                                                            'iyz': {'kilogram':1., 'meter':-2. },
                                                            'izz': {'kilogram':1., 'meter':-2. }}
        transform_dict                                  = {'translation': {'meter': 1.},
                                                            'rotation': {'quaternion':1.},
                                                            'stamp':{'second':1.}}
    # GEOMETRY MSGS   http://docs.ros.org/api/geometry_msgs/html/index-msg.html
        
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  ACCEL
        self.ros_unit_dictionary['geometry_msgs::Accel'] = accel_dict
        self.ros_unit_dictionary['geometry_msgs::AccelStamped'] = copy.deepcopy(accel_dict)
        self.ros_unit_dictionary['geometry_msgs::AccelStamped']['stamp'] = {'second':1.}
        self.ros_unit_dictionary['geometry_msgs::AccelWithCovariance'] = accel_dict
        self.ros_unit_dictionary['geometry_msgs::AccelWithCovarianceStamped'] = copy.deepcopy(accel_dict)
        self.ros_unit_dictionary['geometry_msgs::AccelWithCovarianceStamped']['stamp'] = {'second':1.}

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  INERTIA
        self.ros_unit_dictionary['geometry_msgs::Inertia'] = inertia_dict
        self.ros_unit_dictionary['geometry_msgs::InertiaStamped'] = inertia_dict

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POINT
        self.ros_unit_dictionary['geometry_msgs::Point'] =       {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.}}
        self.ros_unit_dictionary['std::vector<geometry_msgs::Point>'] =       {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POINT
        self.ros_unit_dictionary['geometry_msgs::Point32'] =       {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POINT
        self.ros_unit_dictionary['geometry_msgs::PointStamped'] = {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.},
                                                                    'stamp':{'second':1.}}
        self.ros_unit_dictionary['geometry_msgs::PointStampedPtr'] = {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.},
                                                                    'stamp':{'second':1.}}
        self.ros_unit_dictionary['MessageFilter<geometry_msgs::PointStamped>'] = {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.},
                                                                    'stamp':{'second':1.}}
        self.ros_unit_dictionary['std::vector<geometry_msgs::PointStamped>'] = {'x': {'meter': 1.},
                                                                  'y': {'meter':1.},
                                                                  'z': {'meter':1.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POLYGON
        self.ros_unit_dictionary['geometry_msgs::Polygon'] = {'points': {'meter': 1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POLYGONSTAMPED
        self.ros_unit_dictionary['geometry_msgs::PolygonStamped'] = {'points': {'meter': 1.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POSE
        self.ros_unit_dictionary['geometry_msgs::Pose'] =       {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.}}
        self.ros_unit_dictionary['std::Vector<geometry_msgs::Pose>'] =       {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.}}

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POSE
        self.ros_unit_dictionary['geometry_msgs::Pose2D'] =       {'x': {'meter': 1.},
                                                                  'y': {'meter':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POSESTAMPED
        self.ros_unit_dictionary['geometry_msgs::PoseStamped'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.},
                                                                    'stamp':{'second':1.}}
        self.ros_unit_dictionary['std::Vector<geometry_msgs::PoseStamped>'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POSESTAMPED
        self.ros_unit_dictionary['geometry_msgs::PoseWithCovariance'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.}}
        self.ros_unit_dictionary['geometry_msgs::PoseWithCovariance::_covariance_type'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.}}
        self.ros_unit_dictionary['geometry_msgs::PoseWithCovarianceStamped'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POSE ARRAY  todo
        #self.ros_unit_dictionary['geometry_msgs::PoseArray'] = {'position': {'meter': 1.},
                                                                  #'orientation': {'quaternion':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  Quaternion
        self.ros_unit_dictionary['geometry_msgs::Quaternion'] =       {'x': {'quaternion': 1.},
                                                                  'y': {'quaternion':1.},
                                                                  'z': {'quaternion':1.},
                                                                  'w': {'quaternion':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  QUATERNION STAMPED
        self.ros_unit_dictionary['geometry_msgs::QuaternionStamped'] =       {'x': {'quaternion': 1.},
                                                                  'y': {'quaternion':1.},
                                                                  'z': {'quaternion':1.},
                                                                  'w': {'quaternion':1.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  TRANSFORM
        self.ros_unit_dictionary['geometry_msgs::Transform'] = transform_dict
        self.ros_unit_dictionary['geometry_msgs::TransformStamped'] = transform_dict
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  TWIST
        self.ros_unit_dictionary['geometry_msgs::Twist'] = twist_dict
        self.ros_unit_dictionary['geometry_msgs::TwistStamped'] = copy.deepcopy(twist_dict)
        self.ros_unit_dictionary['geometry_msgs::TwistStamped']['stamp'] = {'second':1.}
        self.ros_unit_dictionary['geometry_msgs::TwistWithCovariance'] = copy.deepcopy(twist_dict)
        self.ros_unit_dictionary['geometry_msgs::TwistWithCovarianceStamped'] = copy.deepcopy(twist_dict)
        self.ros_unit_dictionary['geometry_msgs::TwistWithCovarianceStamped']['stamp'] = {'second':1.}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  WRENCH
        self.ros_unit_dictionary['geometry_msgs::Wrench'] = wrench_dict
        self.ros_unit_dictionary['geometry_msgs::WrenchStamped'] = copy.deepcopy(wrench_dict)
        self.ros_unit_dictionary['geometry_msgs::WrenchStamped']['stamp'] = {'second':1.}


    # NAVIGATION MSGS       http://docs.ros.org/api/nav_msgs/html/index-msg.html
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  GRID CELLS
        self.ros_unit_dictionary['nav_msgs::GridCells'] = {'cell_width': {'meter': 1.}, 
                                                                  'cell_height': {'meter':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  MAP META DATA
        self.ros_unit_dictionary['nav_msgs::MapMetaData'] = {'map_load_time': {'second': 1.}} 
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  ODOMETRY
        self.ros_unit_dictionary['nav_msgs::Odometry'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.},
                                                                    'stamp':{'second':1.},
                                                                    'linear':{'meter':1.,'second':-1.},
                                                                    'angular':{'second':-1.}}
        self.ros_unit_dictionary['nav_msgs::Plan'] = {'position': {'meter': 1.},
                                                                  'orientation': {'quaternion':1.},
                                                                    'stamp':{'second':1.}}


    # SENSOR MSGS   http://docs.ros.org/api/sensor_msgs/html/index-msg.html
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  BATTERY   todo: not implemented in Indigo
        self.ros_unit_dictionary['sensor_msgs::BatteryState'] = {'voltage': {'volt': 1.},
                                                             'cell_voltage': {'volt': 1.},
                                                             'current': {'amp': 1., 'second':1.},
                                                             'capacity': {'amp': 1., 'second':1.},
                                                             'design_capacity': {'amp': 1., 'second':1.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  FLUID PRESSURE 
        self.ros_unit_dictionary['sensor_msgs::FluidPressure'] = {'fluid_pressure': {'pascal': 1.},
                                                                 'variance': {'pascal': 2.},
                                                                 'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  ILLUMINANCE
        self.ros_unit_dictionary['sensor_msgs::Illuminance'] = {'illuminance': {'lux': 1.},
                                                                 'variance': {'lux': 2.},
                                                                 'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  IMU
        self.ros_unit_dictionary['sensor_msgs::Imu'] =      {'angular_velocity': {'second': -1.},
                                                             'angular_velocity_covariance': {'second': -2.},
                                                             'linear_acceleration': {'meter': 1., 'second':-2.},
                                                             'linear_acceleration_covariance': {'meter': 2., 'second':-4.},
                                                             'orientation': {'quaternion': 1.},
                                                             'orientation_covariance': {'quaternion': 2.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  JOINT STATE  #todo: Improve on revolute assumption
        self.ros_unit_dictionary['sensor_msgs::JointState'] = { 'velocity': {'second': -1.},
                                                                'effort': {'kilogram': 1., 'meter':2., 'second':-2.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  LASER ECHO
        self.ros_unit_dictionary['sensor_msgs::LaserEcho'] = {'echoes': {'meter': 1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  LASERSCAN
        self.ros_unit_dictionary['sensor_msgs::LaserScan'] = {'time_increment': {'second': 1.},
                                                             'scan_time': {'second': 1.},
                                                             'range_min': {'meter': 1.},
                                                             'range_max': {'meter': 1.},
                                                             'ranges': {'meter': 1.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  MAGNETIC FIELD
        self.ros_unit_dictionary['sensor_msgs::MagneticField'] = {'magnetic_field': {'tesla': 1.},
                                                                    'magnetic_field_covariance': {'tesla': 2.},
                                                                    'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  MULTI DOF JOINT STATE
        self.ros_unit_dictionary['sensor_msgs::MultiDOFJointState'] = {'twist': twist_dict,
                                                                        'wrench': wrench_dict,
                                                                        'transforms': transform_dict,
                                                                         'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  MULTI ECHO LASERSCAN
        self.ros_unit_dictionary['sensor_msgs::MultiEchoLaserScan'] = {'time_increment': {'second': 1.},
                                                             'scan_time': {'second': 1.},
                                                             'range_min': {'meter': 1.},
                                                             'range_max': {'meter': 1.},
                                                             'ranges': {'meter': 1.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  NAV SAT FIX
        self.ros_unit_dictionary['sensor_msgs::NavSatFix'] = {'latitude': {'degree_360': 1.},
                                                             'longitude': {'degree_360': 1.},
                                                             'altitude': {'meter': 1.},
                                                             'position_covariance': {'meter': 2.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  POINT CLOUD
        self.ros_unit_dictionary['sensor_msgs::PointCloud'] = {'points': {'meter': 1.},
                                                             'stamp':{'second':1.}}
        self.ros_unit_dictionary['sensor_msgs::PointCloud2'] = {'points': {'meter': 1.},
                                                             'stamp':{'second':1.}}
        self.ros_unit_dictionary['sensor_msgs::PointCloud2Iterator<float>'] = {'points': {'meter': 1.},
                                                             'stamp':{'second':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  RANGE
        self.ros_unit_dictionary['sensor_msgs::Range'] =  {'min_range': {'meter':1.},
                                                                'max_range': {'meter':1.},
                                                                'range': {'meter':1.},
                                                                'stamp':{'second':1.},
                                                                'field_of_view':{'radian':1.}}
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  TEMPERATURE
        self.ros_unit_dictionary['sensor_msgs::Temperature'] = {'temperature': {'degree_celsius': 1.},
                                                                'stamp':{'second':1.}}



    # STEREO MSGS       http://docs.ros.org/api/stereo_msgs/html/index-msg.html
        self.ros_unit_dictionary['stereo_msgs::DisparityImage'] =  {'T': {'meter': 1.}, 
                                                                'min_disparity': {'meter':1.},
                                                                'delta_d': {'meter':1.},
                                                                'max_disparity': {'meter':1.}}
    # TRANSFORM MSGS
        self.ros_unit_dictionary['StampedTransform'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf::Quaternion'] =     {'getX': {'quaternion': 1.}, 
                                                            'getY': {'quaternion': 1.},
                                                            'getZ': {'quaternion': 1.},
                                                            'getW': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf::Stamped<tf::Quaternion>'] =     {'getX': {'quaternion': 1.}, 
                                                            'getY': {'quaternion': 1.},
                                                            'getZ': {'quaternion': 1.},
                                                            'getW': {'quaternion': 1.}}
        if self.SHOULD_GIVE_VECTOR3_METERS:
            self.ros_unit_dictionary['tf::Vector3'] =       {'getX': {'meter': 1.}, 
                                                                'getY': {'meter': 1.},
                                                                'getZ': {'meter': 1.},
                                                                 'x': {'meter':1.},
                                                                 'y': {'meter':1.},
                                                                 'z': {'meter':1.}}
            self.ros_unit_dictionary['tf::Stamped<tf::Vector3>'] =       {'getX': {'meter': 1.}, 
                                                                'getY': {'meter': 1.},
                                                                'getZ': {'meter': 1.},
                                                                 'x': {'meter':1.},
                                                                 'y': {'meter':1.},
                                                                 'z': {'meter':1.}}
        self.ros_unit_dictionary['tf::Pose'] =          {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf::Stamped<tf::Pose>'] =          {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf::Transform'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf::StampedTransform'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Stamped<tf::Transform>'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Quaternion'] =       {'getX': {'quaternion': 1.}, 
                                                            'getY': {'quaternion': 1.},
                                                            'getZ': {'quaternion': 1.},
                                                            'getW': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Stamped<tf2::Quaternion>'] =       {'getX': {'quaternion': 1.}, 
                                                            'getY': {'quaternion': 1.},
                                                            'getZ': {'quaternion': 1.},
                                                            'getW': {'quaternion': 1.}}
        if self.SHOULD_GIVE_VECTOR3_METERS:
            self.ros_unit_dictionary['tf2::Vector3'] =       {'getX': {'meter': 1.}, 
                                                                'getY': {'meter': 1.},
                                                                'getZ': {'meter': 1.}}
            self.ros_unit_dictionary['tf2::Stamped<tf2::Vector3>'] =       {'getX': {'meter': 1.}, 
                                                                'getY': {'meter': 1.},
                                                                'getZ': {'meter': 1.}}
        self.ros_unit_dictionary['tf2::Pose'] =          {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Stamped<tf2::Pose>'] =          {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Transform'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::StampedTransform'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}
        self.ros_unit_dictionary['tf2::Stamped<tf2::Transform>'] =       {'getOrigin': {'meter': 1.}, 
                                                            'getRotation': {'quaternion': 1.}}


    # ROS TIME
        if not self.should_ignore_time_and_math:
            self.ros_unit_dictionary['ros::Time'] =  {'secs': {'second': 1.}, 
                                                     'nsec': {'second': 1.}}  # ALL TIME UNITS ARE SECONDS REGARDLESS OF SCALE
            self.ros_unit_dictionary['ros::Duration'] =  {'secs': {'second': 1.}, 
                                                     'nsec': {'second': 1.}}  # ALL TIME UNITS ARE SECONDS REGARDLESS OF SCALE
            self.ros_unit_dictionary['std::vector<ros::Duration>'] =  {'secs': {'second': 1.}, 
                                                     'nsec': {'second': 1.}}  # ALL TIME UNITS ARE SECONDS REGARDLESS OF SCALE
    # KNOWN FUNCTIONS
        if not self.should_ignore_time_and_math:
            self.ros_unit_dictionary['atan2'] =  {'atan2': {'radian': 1.}} 
            self.ros_unit_dictionary['acos'] =  {'acos': {'radian': 1.}} 
            self.ros_unit_dictionary['asin'] =  {'asin': {'radian': 1.}} 
            self.ros_unit_dictionary['atan'] =  {'atan': {'radian': 1.}} 
            self.ros_unit_dictionary['toSec'] =  {'toSec': {'second': 1.}} 
        # self.ros_unit_dictionary['getX'] =  {'getX': {'meter': 1.}} 
        # self.ros_unit_dictionary['getY'] =  {'getY': {'meter': 1.}} 
        # self.ros_unit_dictionary['getZ'] =  {'getZ': {'meter': 1.}} 
        self.ros_unit_dictionary['quatToRPY'] =  {'quatToRPY': {'radian': 1.}} 
    # KNOWN SYMBOL
        # self.ros_unit_dictionary['M_PI'] =  {'M_PI': {'radian': 1}} 
        if self.should_use_dt_heuristic:
            self.ros_unit_dictionary['dt'] =  {'dt': {'second': 1}} 


            

    def find_variable_type(self, variable):
        ''' input: cppcheckdata variable object
            output: string of variable type  (ex:  'int'  or 'std::vector')
            '''
        my_return_string = variable.typeStartToken.str

        if variable.typeStartToken != variable.typeEndToken:
            nextToken = variable.typeStartToken.next
            hasNext = True
            while(hasNext and nextToken):  # BREAK at end of type region
                my_return_string += nextToken.str
                if nextToken.Id == variable.typeEndTokenId:
                    hasNext = False;
                nextToken = nextToken.next
        return my_return_string

    #    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    #       old method - matches variable names by string myVar.linear.x  matched on 'linear'
    #    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 



    def initialize_list_of_vetted_units(self):
        ''' INITIALIZES A LIST OF UNIT DICTIONARIES USED TO TEST UNIT SMELL
            input: none
            output:  none. modified a class level data structure as a side effect
            '''

        self.list_of_vetted_units = [ 
	    {'amp': 1},
	    {'degree_360': 1},
	    {'degree_celsius': 1},
            {'kilogram': 1},
	    {'meter': 1.0},
	    {'meter': 1},
	    {'meter': 2.0},
	    {'meter': 2},
	    {'pascal': 1},
	    {'quaternion': 1},
	    {'quaternion': 2},
	    {'radian': -0.5, 'meter': 0.5},
	    {'radian': -1, 'meter': 1.0},
	    {'radian': -1, 'meter': 1},
	    {'radian': -1.0, 'meter': 1},
	    {'radian': -1},
	    {'radian': -2.0},
	    {'radian': -2},
	    {'radian': 0.5, 'meter': -0.5},
	    {'radian': 1, 'meter': -1.0},
	    {'radian': 1, 'meter': -1},
	    {'radian': 1.0},
	    {'radian': 1},
	    {'radian': 2.0},
	    {'radian': 2},
	    {'second': -1, 'meter': 1.0},
	    {'second': -1, 'meter': 1},
	    {'second': -1, 'radian': 1},
	    {'second': -1.0, 'meter': 1.0},
	    {'second': -1.0, 'radian': 1.0},
	    {'second': -1},
	    {'second': -1.0},
	    {'second': -2, 'kilogram': 1, 'meter': 1},
	    {'second': -2, 'kilogram': 1, 'meter': 2},
	    {'second': -2, 'meter': 1, 'kilogram': 1},
	    {'second': -2, 'meter': 1},
	    {'second': -2, 'meter': 2},
	    {'second': -2, 'meter': 2, 'kilogram': 1},
	    {'second': -2, 'radian': 1, 'kilogram': 1},
	    {'second': -2, 'radian': 1, 'meter': 1},
	    {'second': -2, 'radian': 1},
	    {'second': -2, 'radian': 2, 'kilogram': 1},
	    {'second': -2, 'radian': 2},
	    {'second': -2.0, 'kilogram': 1.0, 'meter': 1.0},
	    {'second': -2.0, 'kilogram': 1.0, 'meter': 2.0},
	    {'second': -2.0, 'meter': 2.0, 'kilogram': 1.0},
	    {'second': -2.0},
	    {'second': -2},
	    {'second': -4, 'meter': 2},
	    {'second': 1, 'meter': -1.0},
	    {'second': 1, 'meter': -1},
	    {'second': 1, 'meter': 1.0},
	    {'second': 1.0, 'meter': -1.0},
	    {'second': 1.0, 'radian': -1.0},
	    {'second': 1},
	    {'second': 2, 'radian': 2, 'meter': -2},
	    {'second': 2.0, 'meter': -1.0},
	    {'second': 2.0},
	    {'second': 2},
	    {'second': 4.0},
	    {'tesla': 1},
	    {'tesla': 2},
            {'volt': 1}
	     ]








#!/usr/bin/env python
"""  COPYRIGHT (c) 2016 UNIVERSITY OF NEBRASKA NIMBUS LAB - JOHN-PAUL ORE
        PHRIKY-UNITS, THE PHYSICAL UNITS INCONSISTENCY DETECTION TOOL
"""

from cps_units_checker import CPSUnitsChecker
from distutils import spawn
from subprocess import Popen
import click
import os
import pkg_resources
from pkg_resources import resource_string
import re
import sys



@click.command()
@click.argument('target_file') 
@click.option('--include_dir', default='', help='include directory for cppcheck')
def main(target_file, include_dir):
    """Console script for phriky_units"""
    print pkg_resources.resource_listdir('phriky_units.resources', '')
    with open('delete_me.txt', 'w') as fp:
        fp.write(resource_string('phriky_units.resources.cppcheck', 'std.cfg'))
    



if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""  COPYRIGHT (c) 2017 UNIVERSITY OF NEBRASKA NIMBUS LAB - JOHN-PAUL ORE
        PHRIKY-UNITS, THE PHYSICAL UNITS INCONSISTENCY DETECTION TOOL 
"""

from __future__ import print_function
from cps_units_checker import CPSUnitsChecker
from distutils import spawn
from subprocess import Popen
import argparse
import os
import pkg_resources
import re
import sys
from shutil import copyfile



def eprint(*args, **kwargs):
    """ HELPER FUNCTION FUNCTION FOR DEBUG PRINTING
        input: <standard python command line input>
        returns: none.  side effect prints to stdout
    """
    print(*args, file=sys.stderr, **kwargs)


def main(
        ):
    """Console script for phriky_units"""
    # 'Detect physical unit inconsistencies in C++ code, especially ROS code.')

    # PARSE COMMAND LINE ARGS
    parser = argparse.ArgumentParser(description='Apply physical unit inconsistency detection to C++ files for ROS.',
                epilog='Copyright 2017 NIMBUS LAB - University of Nebraska, Lincoln, USA')
    parser.add_argument('target_cpp_file', type=str, help='C++ file to be analyzed.')
    parser.add_argument('--include-dir', type=str, help='Include directory.')
    parser.set_defaults(include_dir='')
    parser.add_argument('--database-analysis-audit-id', type=int, help='Database analysis audit id.')
    parser.set_defaults(database_analysis_audit_id=0)
    parser.add_argument('--current-repository', type=str, help='For Database analysis audit trail - sets the Current repository name')
    # DEBUG PRINT AST
    parser.add_argument('--debug-print-ast', dest='debug_print_ast', action='store_true')
    parser.add_argument('--no-debug-print-ast', dest='debug_print_ast', action='store_false')
    parser.set_defaults(debug_print_ast=False)
    parser.add_argument('--show-high-confidence', dest='show_high_confidence', action='store_true')
    parser.add_argument('--no-show-high-confidence', dest='show_high_confidence', action='store_false')
    parser.set_defaults(show_high_confidence=True)
    # SHOW_LOW_CONFIDENCE    
    parser.add_argument('--show-low-confidence', dest='show_low_confidence', action='store_true')
    parser.add_argument('--no-show-low-confidence', dest='show_low_confidence', action='store_false')
    parser.set_defaults(show_low_confidence=False)
    # USE_VAR_NAME_HEURISTIC    
    parser.add_argument('--use-var-name-heuristic', dest='use_var_name_heuristic', action='store_true')
    parser.add_argument('--no-use-var-name-heuristic', dest='use_var_name_heuristic', action='store_false')
    parser.set_defaults(use_var_name_heuristic=False)
    # CHECK_UNIT_SMELL    
    parser.add_argument('--check-unit-smell', dest='check_unit_smell', action='store_true')
    parser.add_argument('--no-check-unit-smell', dest='check_unit_smell', action='store_false')
    parser.set_defaults(check_unit_smell=False)
    # DELETE_DUMP_FILE    
    parser.add_argument('--delete-dump-file', dest='delete_dump_file', action='store_true')
    parser.add_argument('--no-delete-dump-file', dest='delete_dump_file', action='store_false')
    parser.set_defaults(delete_dump_file=True)
    # USE_EXISTING_DUMP_FILE    
    parser.add_argument('--use-existing-dump-file', dest='use_existing_dump_file', action='store_true')
    parser.add_argument('--no-use-existing-dump-file', dest='use_existing_dump_file', action='store_false')
    parser.set_defaults(use_existing_dump_file=True)
    # ONLY_FIND_FILES_WITH_UNITS    
    parser.add_argument('--only-find-files-with-units', dest='only_find_files_with_units', action='store_true')
    parser.add_argument('--no-only-find-files-with-units', dest='only_find_files_with_units', action='store_false')
    parser.set_defaults(only_find_files_with_units=False)
    # WRITE_RESULTS_TO_DATABASE    
    parser.add_argument('--write-results-to-database', dest='write_results_to_database', action='store_true')
    parser.add_argument('--no-write-results-to-database', dest='write_results_to_database', action='store_false')
    parser.set_defaults(write_results_to_database=False)

    # PARSE COMMAND LINE ARGUMENTS
    args = parser.parse_args()

    # target_cpp_file = args.target_cpp_file
    # include_dir  = args.include_dir 
    # debug_print_ast = args.debug_print_ast
    # show_high_confidence = args.show_high_confidence
    # show_low_confidence = args.show_low_confidence
    # use_var_name_heuristic = args.use_var_name_heuristic
    # check_unit_smell = args.check_unit_smell
    # delete_dump_file = args.delete_dump_file
    # use_existing_dump_file = args.use_existing_dump_file
    # only_find_files_with_units = args.only_find_files_with_units
    # write_results_to_database = args.write_results_to_database
    # database_analysis_audit_id = args.database_analysis_audit_id
    # current_repository = args.current_repository
    
    original_directory = os.getcwd()
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    # TEST FOR CPPCHECK
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

    if not spawn.find_executable('cppcheck'):
        # CPPCHECK NOT GLOBALLY INSTALLED, CHECK BIN DIRECTORY
        if not os.path.exists('bin/cppcheck'):
            eprint( 'Could not find required program Cppcheck') #todo
            eprint( 'two options: ')
            eprint( '  1.  sudo apt-get install cppcheck')  #todo
            eprint( '  2.  brew install cppcheck')  #todo
        sys.exit(1)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    # RUN CPPCHECK
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    # EXTRACT DIR

    if not os.path.exists(args.target_cpp_file):
        eprint( 'file does not exist: %s' % args.target_cpp_file)
        sys.exit(1)

    # print( 'Run cppcheck...'),
    target_cpp_file_dir = os.path.dirname(args.target_cpp_file)
    target_cpp_file  = args.target_cpp_file.replace("//", "/")
    target_cpp_file_base_name = os.path.basename(target_cpp_file)
    dump_filename = os.path.basename(target_cpp_file) + '.dump'

    # CHECK IF DUMP EXISTS, AND IF SO, SHOULD WE USE IT
    if not os.path.exists(target_cpp_file + '.dump' ) or not args.use_existing_dump_file:
        # if not os.path.exists(dump_filename):
            # print("does not exist : %s" % (target_cpp_file + '.dump'))
        # else:
            # print("exists! : %s " % (target_cpp_file + '.dump'))
                
        os.chdir(target_cpp_file_dir)
        # CREATE LOCAL COPY OF CFG
        if pkg_resources.resource_exists('phriky_units', 'resources'):
            if pkg_resources.resource_exists('phriky_units', 'resources/cppcheck'):
                if pkg_resources.resource_exists('phriky_units', 'resources/cppcheck/std.cfg'):
                    if not os.path.exists('cfg'):
                        os.makedirs('cfg')
                    with open('std.cfg', 'w') as f:
                        f.write(pkg_resources.resource_string('phriky_units', 'resources/cppcheck/std.cfg'))
                    os.rename('std.cfg', 'cfg/std.cfg')
        else:
            eprint("resource phriky_units not found: trying local option")
            if not os.path.exists('cfg'):
                os.makedirs('cfg')
            copyfile(os.path.join(original_directory, 'std.cfg'), os.path.join(os.path.join(target_cpp_file_dir, 'cfg'), 'std.cfg'))
        if not os.path.exists(dump_filename):
            args = ['cppcheck', '--dump', '-I ../include', '--suppress="*"', '-q', target_cpp_file_base_name]
            cppcheck_process = Popen(' '.join(args),  shell=True)
            cppcheck_process.communicate()
            if cppcheck_process.returncode != 0:
                eprint( 'cppcheck appears to have failed...exiting with return code %d' % cppcheck_process.returncode)
                sys.exit(1)
            eprint( "Created: %s'" % dump_filename)
        # REMOVE LOCAL COPY OF CFG
        if os.path.exists('cfg/std.cfg'):
            try:
                os.remove('cfg/std.cfg')
                os.rmdir('cfg')
            except:
                # eprint('problem removing cfg folder')
                pass # todo - fail silently for now
         
        # RETURN TO HOME
        os.chdir(original_directory)

    print( 'TARGET: %s' % target_cpp_file )
    cps_unit_checker = CPSUnitsChecker()
    dump_file = os.path.join(os.path.dirname(target_cpp_file), dump_filename)
    source_file = dump_file.replace('.dump','')
    #   APPLY COMMAND LINE CONFIGURATION OPTIONS
    cps_unit_checker.debug_print_AST = args.debug_print_ast
    cps_unit_checker.SHOULD_MATCH_ON_HEURISTIC_VARIABLE_NAMES = args.use_var_name_heuristic
    cps_unit_checker.SHOULD_ONLY_FIND_FILES_WITH_UNITS = args.only_find_files_with_units
    cps_unit_checker.SHOULD_CHECK_UNIT_SMELLS = args.check_unit_smell
    if args.write_results_to_database:
        cps_unit_checker.SHOULD_WRITE_RESULTS_TO_DATABASE = True  # SHOULD BE REDUNDANT
        # ACTUALLY INITIALIZES THE DATABASE CONNECTION 
        cps_unit_checker.init_database_connection()
        # ATTEMPTS TO GET GIT REPO NAME
        cps_unit_checker.database_connector.database_analysis_audit_id = args.database_analysis_audit_id
        cps_unit_checker.database_connector.current_repository = args.current_repository
        cps_unit_checker.database_connector.current_file_under_analysis = target_cpp_file
    cps_unit_checker.main_run_check(dump_file, source_file)
    # OVERRIDE DELETING THE DUMP FILE WHEN IT EXISTS
    if args.use_existing_dump_file:
        args.delete_dump_file = False

    # COMMAND LINE OPTION TO KEEP THE DUMP FILE
    if args.delete_dump_file or not cps_unit_checker.found_ros_units_in_this_file:
        try:
            os.remove(dump_file)
            print("removing %s" % dump_file)
        except: 
            pass


    # for e in cps_unit_checker.errors:
    if len(cps_unit_checker.error_checker.all_errors) > 0:
        print ("inconsistencies for file: %s" % target_cpp_file)
        cps_unit_checker.error_checker.pretty_print(args.show_high_confidence, args.show_low_confidence)
        # eprint( "%s" % e.get_error_desc())
        # eprint( "%s" % e.linenr)
        # eprint( "%s" % e.units_at_first_assignment)
        # eprint( "%s" % e.all_units_assigned_to_var)
        # eprint( "%s" % e.units_when_multiple_happened)
    # eprint("Total errors: %i" % len(cps_unit_checker.errors))




if __name__ == "__main__":
    main()
